#ifndef GPIOAccess_H
#define GPIOAccess_H

#include "GPIOTypes.h"
#include "GPIOPwmPin.h"

#define numPins 27

class GPIOAccess {
public:
    static GPIO_Result setDirection(int pinNum, GPIO_Direction dir);
    static GPIO_Result getDirection(int pinNum, GPIO_Direction &dir);
    static GPIO_Result set(int pinNum, int value);
    static GPIO_Result get(int pinNum, int &value);
    
    static GPIO_Result setPWM(int pinNum, int freq, int duty);
    static GPIO_Result startPWM(int pinNum);
    static GPIO_Result stopPWM(int pinNum);
    
    static int getPWMFreq(int pinNum);
    static int getPWMDuty(int pinNum);
    static bool isPWMRunning(int pinNum);
    
    static bool isPinUsable(int pinNum);
    static bool isAccessOk();
    
private:
    static GPIO_Result checkAndSetupAddress(unsigned long int blockBaseAddr, unsigned long int blockSize);
    static GPIO_Result checkPinAndAccess(int pin);

    static GPIO_Direction getDirectionImpl(int pinNum);
    
    static void writeReg(unsigned long int registerOffset, unsigned long int value);
    static unsigned long int readReg(unsigned long int registerOffset);

    static void setBit(unsigned long int &regVal, int bitNum, int value);
    static int getBit(unsigned long int regVal, int bitNum);

    static volatile unsigned long int *regAddress;
    static bool pinUsable[numPins];
    static GPIOPwmPin * pwmPin[numPins];
    static bool pwmSetup;
};

#endif
